package de.hft.stuttgart.citydoctor2.datastructure;

import de.hft.stuttgart.citydoctor2.check.Check;
import de.hft.stuttgart.citydoctor2.check.CheckError;
import de.hft.stuttgart.citydoctor2.check.CheckId;
import de.hft.stuttgart.citydoctor2.parser.ParserConfiguration;
import de.hft.stuttgart.citydoctor2.utils.CopyHandler;
import de.hft.stuttgart.citydoctor2.utils.Copyable;
import org.citygml4j.core.model.deprecated.transportation.TransportationComplex;
import org.citygml4j.core.model.transportation.Railway;
import org.citygml4j.core.model.transportation.Road;
import org.citygml4j.core.model.transportation.Square;
import org.citygml4j.core.model.transportation.Track;
import org.citygml4j.core.model.transportation.Waterway;
import org.citygml4j.core.util.geometry.GeometryFactory;

import java.util.ArrayList;
import java.util.List;

public class TopLevelTransportFeature extends TransportationSpace {

    private final List<TransportSection> sections = new ArrayList<>();
    private final List<TransportSection> intersections = new ArrayList<>();

    public static TopLevelTransportFeature from(Track t) {
        TopLevelTransportFeature top = new TopLevelTransportFeature(TransportationType.TRACK);
        top.setGmlObject(t);
        return top;
    }

    public static TopLevelTransportFeature from(Road r) {
        TopLevelTransportFeature top = new TopLevelTransportFeature(TransportationType.ROAD);
        top.setGmlObject(r);
        return top;
    }

    public static TopLevelTransportFeature from(Waterway w) {
        TopLevelTransportFeature top = new TopLevelTransportFeature(TransportationType.WATERWAY);
        top.setGmlObject(w);
        return top;
    }

    public static TopLevelTransportFeature from(Railway r) {
        TopLevelTransportFeature top = new TopLevelTransportFeature(TransportationType.RAILWAY);
        top.setGmlObject(r);
        return top;
    }

    public static TopLevelTransportFeature from(Square s) {
        TopLevelTransportFeature top = new TopLevelTransportFeature(TransportationType.SQUARE);
        top.setGmlObject(s);
        return top;
    }

    public static TopLevelTransportFeature from(TransportationComplex tc) {
        TopLevelTransportFeature top = new TopLevelTransportFeature(TransportationType.TRANSPORTATION_COMPLEX);
        top.setGmlObject(tc);
        return top;
    }

    private TopLevelTransportFeature(TransportationType type) {
        super(type);
    }


    public void addSection(TransportSection section) {
        sections.add(section);
    }

    public void addIntersection(TransportSection section) {
        intersections.add(section);
    }

    public List<TransportSection> getSections() {
        return sections;
    }

    public List<TransportSection> getIntersections() {
        return intersections;
    }

    @Override
    public void reCreateGeometries(GeometryFactory factory, ParserConfiguration config) {
        super.reCreateGeometries(factory, config);
        for (TransportSection section : sections) {
            section.reCreateGeometries(factory, config);
        }
        for (TransportSection section : intersections) {
            section.reCreateGeometries(factory, config);
        }
    }


    @Override
    public boolean containsError(CheckId checkIdentifier) {
        boolean hasError = super.containsError(checkIdentifier);
        if (hasError) {
            return true;
        }
        for (TransportSection section : sections) {
            if (section.containsError(checkIdentifier)) {
                return true;
            }
        }
        for (TransportSection section : intersections) {
            if (section.containsError(checkIdentifier)) {
                return true;
            }
        }
        return false;
    }

    @Override
    public void clearAllContainedCheckResults() {
        super.clearAllContainedCheckResults();
        for (TransportSection section : sections) {
            section.clearAllContainedCheckResults();
        }
        for (TransportSection section : intersections) {
            section.clearAllContainedCheckResults();
        }
    }

    @Override
    public void collectContainedErrors(List<CheckError> errors) {
        super.collectContainedErrors(errors);
        for (TransportSection section : sections) {
            section.collectContainedErrors(errors);
        }
        for (TransportSection section : intersections) {
            section.collectContainedErrors(errors);
        }
    }

    @Override
    public boolean containsAnyError() {
        boolean hasError = super.containsAnyError();
        if (hasError) {
            return true;
        }
        for (TransportSection section : sections) {
            if (section.containsAnyError()) {
                return true;
            }
        }
        for (TransportSection section : intersections) {
            if (section.containsAnyError()) {
                return true;
            }
        }
        return false;
    }

    @Override
    public void accept(Check c) {
        super.accept(c);
        if (c.canExecute(this)) {
            c.check(this);
        }
        for (TransportSection section : sections) {
            section.accept(c);
        }
        for (TransportSection section : intersections) {
            section.accept(c);
        }
    }


    @Override
    public void unsetGmlGeometries() {
        super.unsetGmlGeometries();
        for (TransportSection section : sections) {
            section.unsetGmlGeometries();
        }
        for (TransportSection section : intersections) {
            section.unsetGmlGeometries();
        }
    }

    @Override
    public CityObject getTopLevelCityObject() {
        return this;
    }


    @Override
    public String toString() {
        return "TopLevelTransportFeature [id=" + getGmlId() + "]";
    }

    @Override
    public void prepareForChecking() {
        super.prepareForChecking();
        for (TransportSection section : sections) {
            section.prepareForChecking();
        }
        for (TransportSection section : intersections) {
            section.prepareForChecking();
        }
    }

    @Override
    public void clearMetaInformation() {
        super.clearMetaInformation();
        for (TransportSection section : sections) {
            section.clearMetaInformation();
        }
        for (TransportSection section : intersections) {
            section.clearMetaInformation();
        }
    }

    @Override
    public void collectInstances(CopyHandler handler) {
        super.collectInstances(handler);
        for (TransportSection section : sections) {
            section.collectInstances(handler);
        }
        for (TransportSection section : intersections) {
            section.collectInstances(handler);
        }
    }

    @Override
    public void fillValues(Copyable original, CopyHandler handler) {
        super.fillValues(original, handler);
        TransportationObject originalTo = (TransportationObject) original;
        for (TransportSection section : sections) {
            section.fillValues(originalTo, handler);
        }
        for (TransportSection section : intersections) {
            section.fillValues(originalTo, handler);
        }
    }

    @Override
    public Copyable createCopyInstance() {
        return new TopLevelTransportFeature(type);
    }


}
