package de.hft.stuttgart.citydoctor2.tesselation;

import java.util.ArrayList;
import java.util.List;

import de.hft.stuttgart.citydoctor2.datastructure.LinearRing;
import de.hft.stuttgart.citydoctor2.datastructure.Polygon;
import de.hft.stuttgart.citydoctor2.datastructure.Vertex;
import de.hft.stuttgart.citydoctor2.math.ProjectionAxis;
import de.hft.stuttgart.citydoctor2.math.Triangle3d;
import earcut4j.Earcut;

/**
 * Tesselator to create triangles out of polygons. Uses earcut4j implementation.
 * 
 * @author Matthias Betz
 *
 */
public class EarcutTesselator {

	public static TesselatedPolygon tesselatePolygon(Polygon p) {
		// sum number of vertices, don't use duplicate last point
		// assume polygons are closed
		int nrOfVertices = p.getExteriorRing().getVertices().size() - 1;
		int[] holeStart = null;
		if (!p.getInnerRings().isEmpty()) {
			holeStart = new int[p.getInnerRings().size()];
			List<LinearRing> innerRings = p.getInnerRings();
			for (int i = 0; i < innerRings.size(); i++) {
				LinearRing r = innerRings.get(i);
				holeStart[i] = nrOfVertices;
				nrOfVertices += r.getVertices().size() - 1;
			}
		}
		// collect vertices
		// find most dominant projection axis for 2d projection
		ProjectionAxis axis = ProjectionAxis.of(p);
		
		double[] vertices = new double[nrOfVertices * 2];
		List<Vertex> vertexObjects = new ArrayList<>();
		addVerticesToList(p.getExteriorRing(), vertexObjects);
		for (LinearRing innerRing : p.getInnerRings()) {
			addVerticesToList(innerRing, vertexObjects);
		}
		
		// write the vector data according to the projection axis into the array
		int start = addRingToArray(p.getExteriorRing(), vertices, 0, axis);
		for (LinearRing innerRing : p.getInnerRings()) {
			start = addRingToArray(innerRing, vertices, start, axis);
		}
		
		// triangulation
		List<Integer> indices = Earcut.earcut(vertices, holeStart, 2);
		List<Triangle3d> triangles = new ArrayList<>();
		if (indices.size() % 3 != 0) {
			throw new IllegalStateException();
		}
		for (int i = 0; i < indices.size(); i = i + 3) {
			Triangle3d t = new Triangle3d(vertexObjects.get(indices.get(i + 0)),
					vertexObjects.get(indices.get(i + 1)),
					vertexObjects.get(indices.get(i + 2)));
			triangles.add(t);
		}
		return new TesselatedPolygon(triangles, p);
	}

	private static void addVerticesToList(LinearRing ring, List<Vertex> vertexObjects) {
		List<Vertex> vertices = ring.getVertices();
		for (int i = 0; i < vertices.size() - 1; i++) {
			Vertex v = vertices.get(i);
			vertexObjects.add(v);
		}
	}

	private static int addRingToArray(LinearRing ring, double[] vertices, int start, ProjectionAxis axis) {
		List<Vertex> ringVertices = ring.getVertices();
		for (int i = 0; i < ringVertices.size() - 1; i++) {
			Vertex v = ringVertices.get(i);
			axis.writeCoordinatesOfVectorInArray(v, vertices, start);
			start = start + 2;
		}
		return start;
	}
	
	private EarcutTesselator() {
	}

}
